from fmtr.tools.import_tools import MissingExtraMockModule
from fmtr.tools.path_tools.path_tools import Path, PackagePaths

try:
    from fmtr.tools.path_tools.app_path_tools import AppPaths
except ModuleNotFoundError as exception:
    AppPaths = MissingExtraMockModule('path.app', exception)

try:
    from fmtr.tools.path_tools.type_path_tools import guess
except ModuleNotFoundError as exception:
    guess = MissingExtraMockModule('path.type', exception)



from datetime import datetime, timezone

MIN = datetime.min.replace(tzinfo=timezone.utc)
MAX = datetime.max.replace(tzinfo=timezone.utc)

def now() -> datetime:
    """

    Now UTC

    """
    return datetime.now(tz=timezone.utc)



from inspect import stack


def get_call_path(offset=1):
    """

    Get the path of the calling module

    """
    from fmtr.tools.path_tools import Path
    frames = stack()
    frame_called = frames[offset]
    path = Path(frame_called.filename).absolute()
    return path



from typing import Any

from fmtr.tools.constants import Constants


class MissingExtraError(ImportError):
    """

    Error to raise if extras are missing.

    """

    MASK = 'The current module is missing dependencies. To install them, run: `pip install {library}[{extra}] --upgrade`'

    def __init__(self, extra):
        self.message = self.MASK.format(library=Constants.LIBRARY_NAME, extra=extra)

        super().__init__(self.message)


def identity(x: Any) -> Any:
    """

    Dummy (identity) function

    """
    return x


class Special:
    """

    Classes to differentiate special arguments from primitive arguments.

    """


class Empty(Special):
    """

    Class to denote an unspecified object (e.g. argument) when `None` cannot be used.

    """


class Raise(Special):
    """

    Class to denote when a function should raise instead of e.g. returning a default.

    """


class Auto(Special):
    """

    Class to denote when an argument should be inferred.

    """


class Required(Special):
    """

    Class to denote when an argument is required.

    """



EMPTY = Empty()



from fmtr.tools.import_tools import MissingExtraMockModule

from fmtr.tools.setup_tools.setup_tools import Setup, SetupPaths, Dependencies, Tools

try:
    from setuptools import find_namespace_packages, find_packages, setup as setup_setuptools
except ModuleNotFoundError as exception:
    find_namespace_packages = find_packages = setup_setuptools = MissingExtraMockModule('setup', exception)



from typing import List, Optional, Any, Annotated, Generic

import pydantic_ai
from pydantic import PlainValidator
from pydantic_ai import RunContext, ModelRetry
from pydantic_ai._output import OutputDataT
from pydantic_ai.agent import AgentRunResult, Agent
from pydantic_ai.models.openai import OpenAIModel
from pydantic_ai.output import OutputSpec, NativeOutput, ToolOutput
from pydantic_ai.providers.openai import OpenAIProvider
from pydantic_ai.tools import AgentDepsT

from fmtr.tools import environment_tools as env
from fmtr.tools.constants import Constants
from fmtr.tools.logging_tools import logger
from fmtr.tools.string_tools import truncate_mid

pydantic_ai.Agent.instrument_all()


class Validator:
    """

    Subclassable validator

    """

    async def validate(self, ctx: RunContext[Any], output: Any) -> List[str]:
        raise NotImplementedError()


api_key = env.get(Constants.FMTR_OPENAI_API_KEY_KEY, None)

class Task(Generic[AgentDepsT, OutputDataT]):
    """

    Linear task definition, as Agent configuration and typing, plus state history.

    """

    TypeDeps = str
    TypeOutput = str

    PROVIDER = OpenAIProvider(api_key=api_key) if api_key else None

    API_HOST_FMTR = env.get(Constants.FMTR_AI_HOST_KEY, Constants.FMTR_AI_HOST_DEFAULT)
    API_URL_FMTR = f'https://{API_HOST_FMTR}/v1'
    PROVIDER_FMTR = OpenAIProvider(base_url=API_URL_FMTR)

    MODEL_ID = 'gpt-4o'
    MODEL_ID_FMTR = 'qwen2.5-coder:32b'
    SYSTEM_PROMPT_STATIC = None
    RESULT_RETRIES = 5
    VALIDATORS: List[Validator] = []

    def __init__(self, *args, **kwargs):
        """

        Configure Agent

        """

        self.model = OpenAIModel(self.MODEL_ID, provider=self.PROVIDER)
        self.agent = Agent[AgentDepsT, OutputDataT](
            *args,
            model=self.model,
            system_prompt=self.SYSTEM_PROMPT_STATIC or [],
            deps_type=self.TypeDeps,
            output_type=self.tool_output,
            output_retries=self.RESULT_RETRIES,
            **kwargs
        )

        self.agent.output_validator(self.validate)
        self.agent.system_prompt(self.add_system_prompt)
        self.history = []

    @property
    def tool_output(self) -> OutputSpec[OutputDataT]:
        """

        Tool output specification (e.g. ToolOutput/NativeOutput etc.)

        """
        return ToolOutput(self.TypeOutput)


    @property
    def sync_runner(self):
        """

        Convenience/debug function to run without async.

        """
        import asyncio
        return asyncio.run

    async def run(self, *args, deps=None, **kwargs) -> AgentRunResult[OutputDataT]:
        """

        Run Agent with deps-relative user prompt and while storing history

        """
        result = await self.agent.run(*args, user_prompt=self.get_prompt(deps), deps=deps, message_history=self.history, **kwargs)
        self.history = result.all_messages()
        return result

    async def validate(self, ctx: RunContext[AgentDepsT], output: OutputDataT) -> OutputDataT:
        """

        Aggregate any validation failures and combine them into a single ModelRetry exception

        """
        msgs = []
        for validator in self.VALIDATORS:
            msgs += validator.validate(ctx, output)

        if msgs:
            msg = '. '.join(msgs)
            logger.warning(msg)
            raise ModelRetry(msg)

        return output

    def get_prompt(self, deps: Optional[AgentDepsT]) -> Optional[str]:
        """

        Dummy prompt generator

        """
        return None

    def add_system_prompt(self, ctx: RunContext[AgentDepsT]) -> str | List[str]:
        """

        Dummy system prompt append

        """

        return []

    def reset(self):
        """

        Reset the task by deleting its history.

        """
        self.history = []

    @property
    def tool_schema(self):
        """

        Impossible to find otherwise.

        """
        return self.agent._output_toolset

    def __repr__(self):
        """

        String representation of the object

        """
        return f'{self.__class__.__name__}({repr(truncate_mid(self.SYSTEM_PROMPT_STATIC, 100))})'


def default_prompt_none_specified(text):
    """

    If the prompt is falsey, explicitly state None Specified

    """
    if not (text or '').strip():
        return Constants.PROMPT_NONE_SPECIFIED
    return text


StringDefaultNoneSpecified = Annotated[Optional[str], PlainValidator(default_prompt_none_specified)]


if __name__ == '__main__':
    import asyncio
    from fmtr.tools import dm


    class TestOutput(dm.Base):
        text: str


    class TestDeps(dm.Base):
        lang: str
        subject: str


    class TaskTest(Task):
        PROVIDER = Task.PROVIDER_FMTR
        MODEL_ID = Task.MODEL_ID_FMTR
        TypeOutput = TestOutput
        SYSTEM_PROMPT_STATIC = 'Tell the user jokes.'

        @property
        def tool_output(self) -> OutputSpec[TestOutput]:
            return NativeOutput(self.TypeOutput)

        def add_system_prompt(self, ctx: RunContext[TestDeps]) -> str:
            return f'The jokes must be in the {ctx.deps.lang} language.'

        def get_prompt(self, deps: Optional[TestDeps]) -> str:
            return f'Tell me one about {deps.subject}.'

    task = TaskTest()
    deps = TestDeps(lang='English', subject='eggs')
    result1 = task.sync_runner(task.run(deps=deps))
    result1

    deps = TestDeps(lang='German', subject='sausages')
    result2 = task.sync_runner(task.run(deps=deps))
    result2



import dataclasses

import pydevd_pycharm

from fmtr.tools import environment_tools as env
from fmtr.tools.constants import Constants

MASK = 'Starting debugger at tcp://{host}:{port}...'


@dataclasses.dataclass
class ShellDebug:
    """

    Debugging information for shell commands

    """
    command: str
    out: str
    err: str
    status: int

    # timestamp: datetime=dataclasses.field(default_factory=datetime.now)

    @classmethod
    def from_path(cls, path_str):
        """

        Get debug info from path

        """
        from fmtr.tools import Path
        path = Path(path_str).absolute()
        data = {field.name: (path / f'{field.name}.log').read_text().strip() for field in dataclasses.fields(cls)}
        self = cls(**data)
        return self

    @property
    def env(self):
        return env.get_dict()



def trace(is_debug=None, host=None, port=None, stdoutToServer=True, stderrToServer=True, **kwargs):
    """

    Connect to PyCharm debugger if enabled

    """
    if not is_debug:
        is_debug = env.get_bool(Constants.FMTR_REMOTE_DEBUG_ENABLED_KEY, False)

    if not is_debug:
        return

    if is_debug is True and not host:
        host = Constants.FMTR_REMOTE_DEBUG_HOST_DEFAULT

    host = host or env.get(Constants.FMTR_REMOTE_DEBUG_HOST_KEY, Constants.FMTR_REMOTE_DEBUG_HOST_DEFAULT)
    port = port or Constants.FMTR_REMOTE_DEBUG_PORT_DEFAULT

    from fmtr.tools import logger

    msg = MASK.format(host=host, port=port)
    logger.info(msg)

    pydevd_pycharm.settrace(host, port=port, stdoutToServer=stdoutToServer, stderrToServer=stderrToServer, **kwargs)


def debug_shell():
    """

    Starts a debug shell by initializing a `ShellDebug` object from a given path
    and enabling tracing with debug mode turned on.

    """
    import sys
    path_str = sys.argv[1]
    data = ShellDebug.from_path(path_str)
    trace(is_debug=True)
    data


if __name__ == "__main__":
    import sys

    sys.argv = [
        'test.py',
        './fmtr-debug/34e8d492-2f15-419a-8fcb-fe4fa0fa02bb',
    ]
    debug_shell()



from functools import lru_cache
from openai import OpenAI

from fmtr.tools import environment_tools

OPENAI_TOKEN_KEY = 'FMTR_OPENAI_API_KEY'


@lru_cache
def get_client():
    """

    Get an OpenAI client instance

    """
    OPENAI_KEY = environment_tools.get(OPENAI_TOKEN_KEY)
    client = OpenAI(api_key=OPENAI_KEY)
    return client


def get_text(prompt, model='gpt-4o'):
    """

    Very simple prompt-to-output

    """
    client = get_client()
    messages = [{"role": 'user', "content": prompt}]

    completion = client.chat.completions.create(
        messages=messages,
        model=model
    )
    text = completion.choices[0].message.content
    return text


if __name__ == '__main__':
    text = get_text('hi')
    text



from dataclasses import dataclass

from fmtr.tools.dns_tools import server, client
from fmtr.tools.dns_tools.dm import Exchange
from fmtr.tools.logging_tools import logger


@dataclass(kw_only=True, eq=False)
class Proxy(server.Plain):
    """

    Base for a DNS Proxy server (plain server) TODO: Allow subclassing of any server type.

    """

    client: client.HTTP

    def process_question(self, exchange: Exchange):
        """

        Modify exchange based on initial question.

        """
        return

    def process_upstream(self, exchange: Exchange):
        """

        Modify exchange after upstream response.

        """
        return

    def finalize(self, exchange: Exchange):
        """

        Finalize a still open exchange.

        """
        exchange.is_complete = True

    async def resolve(self, exchange: Exchange) -> Exchange:
        """

        Resolve a request, processing each stage, initial question, upstream response etc.
        Subclasses can override the relevant processing methods to implement custom behaviour.

        """
        with logger.span(f'Processing question...'):
            self.process_question(exchange)
        if exchange.is_complete:
            return exchange

        with logger.span(f'Making upstream request...'):
            self.client.resolve(exchange)
        if exchange.is_complete:
            return exchange

        with logger.span(f'Processing upstream response...'):
            self.process_upstream(exchange)
        if exchange.is_complete:
            return exchange

        self.finalize(exchange)

        return exchange



import time
from multiprocessing import Queue, Process

from fmtr.tools.logging_tools import logger


class ContextProcess:
    """

    Start/stop a function in a process using a context manager

    """

    def __init__(self, func, restart_delay=5, **kwargs):
        """

        Set up queue etc.

        """
        self.func = func
        self.func_args = kwargs
        self.queue = Queue()
        self.restart_delay = restart_delay
        self.process = None

    def start(self):
        """

        Start processes.

        """

        msg = f'Starting process {self.func.__name__}...'
        logger.warning(msg)

        self.process = Process(target=self.func, kwargs=self.func_args)
        self.process.start()

    def stop(self):
        """

        Stop all processes.

        """

        msg = f'Stopping training processes...'
        logger.warning(msg)
        self.process.terminate()
        self.process = None

    def restart(self):
        """

        Restart the process.

        """
        self.stop()
        time.sleep(self.restart_delay)
        self.start()

    def __enter__(self):
        """

        Start processes in context manager.

        """
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """

        Stop processes when the context manager exits.

        """
        if self.process:
            self.stop()



from contextlib import contextmanager, ExitStack


@contextmanager
def null():
    """

    Null context manager.

    """
    yield


@contextmanager
def contexts(*contexts):
    """

    Tee context managers.

    """
    with ExitStack() as stack:
        resources = [stack.enter_context(context) for context in contexts]
        yield resources



import sys
from datetime import datetime
from fnmatch import fnmatch
from functools import cached_property
from itertools import chain
from typing import List, Dict, Any, Callable, Optional

from fmtr.tools.constants import Constants
from fmtr.tools.path_tools import Path
from fmtr.tools.path_tools.path_tools import FromCallerMixin


class SetupPaths(FromCallerMixin):
    """

    Canonical paths for a repo.

    """

    def __init__(self, path=None, org=Constants.ORG_NAME):
        """

        Use calling module path as default path, if not otherwise specified.

        """
        if not path:
            path = self.from_caller()

        self.org_name = org
        self.repo = Path(path)

    @property
    def readme(self) -> Path:
        """

        Path of the README file.

        """
        return self.repo / 'README.md'

    @property
    def version(self) -> Path:
        """

        Path of the version file

        """
        return self.path / Constants.FILENAME_VERSION

    @cached_property
    def path(self) -> Path:
        """

        Infer the package path. It should be the only non-excluded package in the repo/org Path.

        """

        if self.is_namespace:
            base = self.org
        else:
            base = self.repo

        packages = [
            dir for dir in base.iterdir()
            if (dir / Constants.INIT_FILENAME).is_file()
               and not any(fnmatch(dir.name, pattern) for pattern in Constants.PACKAGE_EXCLUDE_DIRS)
        ]

        if len(packages) != 1:
            dirs_str = ', '.join([str(dir) for dir in packages])
            msg = f'Expected exactly one package in {self.repo}, found {dirs_str}'
            raise ValueError(msg)

        package = next(iter(packages))
        return package

    @property
    def org(self) -> bool | Path:
        """

        Get the org path, i.e. the namespace parent directory.

        """
        if not self.org_name:
            return False
        org = self.repo / self.org_name
        if not org.is_dir():
            return False
        return org

    @property
    def entrypoint(self) -> Path:
        """

        Path of base entrypoint module.

        """
        return self.path / Constants.ENTRYPOINT_FILE

    @property
    def entrypoints(self) -> Path:
        """

        Path of entrypoints sub-package.

        """
        return self.path / Constants.ENTRYPOINTS_DIR

    @property
    def is_namespace(self) -> bool:
        return bool(self.org)

    @property
    def name(self) -> str:
        return self.path.stem


class Setup(FromCallerMixin):
    """

    Abstract canonical pacakge setup for setuptools.

    """
    AUTHOR = 'Frontmatter'
    AUTHOR_EMAIL = 'innovative.fowler@mask.pro.fmtr.dev'

    REQUIREMENTS_ARG = 'requirements'

    ENTRYPOINT_COMMAND_SEP = '-'
    ENTRYPOINT_FUNCTION_SEP = '_'
    ENTRYPOINT_FUNC_NAME = 'main'

    def __init__(self, dependencies, paths=None, org=Constants.ORG_NAME, client=None, do_setup=True, **kwargs):
        """

        First check if commandline arguments for requirements output exist. If so, print them and return early.
        Otherwise, continue generating data to pass to setuptools.

        """
        self.kwargs = kwargs

        if type(dependencies) is not Dependencies:
            dependencies = Dependencies(**dependencies)
        self.dependencies = dependencies

        requirements_extras = self.get_requirements_extras()

        if requirements_extras:
            self.print_requirements()
            return

        self.org = org

        if not paths:
            paths = SetupPaths(path=self.from_caller(), org=self.org)
        self.paths = paths

        self.client = client

        if do_setup:
            self.setup()

    def get_requirements_extras(self) -> Optional[List[str]]:
        """

        Get list of extras from command line arguments.

        """
        if self.REQUIREMENTS_ARG not in sys.argv:
            return None

        extras_str = sys.argv[-1]
        extras = extras_str.split(',')
        return extras

    def print_requirements(self):
        """

        Output flat list of requirements for specified extras

        """
        reqs = []
        reqs += self.dependencies.install

        for extra in sys.argv[-1].split(','):
            reqs += self.dependencies.extras[extra]
        reqs = '\n'.join(reqs)
        print(reqs)

    @property
    def console_scripts(self) -> List[str]:
        """

        Generate console scripts for the `entrypoint` module - and/or any modules in `entrypoints` sub-package.

        """

        if not self.paths.entrypoints.exists():
            paths_mods = []
        else:
            paths_mods = list(self.paths.entrypoints.iterdir())

        names_mods = [path.stem for path in paths_mods if path.is_file() and path.name != Constants.INIT_FILENAME]
        command_suffixes = [name_mod.replace(self.ENTRYPOINT_FUNCTION_SEP, self.ENTRYPOINT_COMMAND_SEP) for name_mod in names_mods]
        commands = [f'{self.name_command}-{command_suffix}' for command_suffix in command_suffixes]
        paths = [f'{self.name}.{Constants.ENTRYPOINTS_DIR}.{name_mod}:{self.ENTRYPOINT_FUNC_NAME}' for name_mod in names_mods]

        if self.paths.entrypoint.exists():
            commands.append(self.name_command)
            path = f'{self.name}.{self.paths.entrypoint.stem}:{self.ENTRYPOINT_FUNC_NAME}'
            paths.append(path)

        console_scripts = [f'{command} = {path}' for command, path in zip(commands, paths)]

        return console_scripts

    @cached_property
    def name_command(self) -> str:
        """

        Name as a command, e.g. `fmtr-tools`

        """
        return self.name.replace('.', self.ENTRYPOINT_COMMAND_SEP)

    @property
    def name(self) -> str:
        """

        Full library name

        """
        if self.paths.is_namespace:
            return f'{self.paths.org_name}.{self.paths.name}'
        return self.paths.name

    @property
    def author(self) -> str:
        """

        Create appropriate author string

        """
        if self.client:
            return f'{self.AUTHOR} on behalf of {self.client}'
        return self.AUTHOR

    @property
    def copyright(self) -> str:
        """

        Create appropriate copyright string

        """
        if self.client:
            return self.client
        return self.AUTHOR

    @property
    def long_description(self) -> str:
        """

        Read in README.md

        """
        return self.paths.readme.read_text()

    @property
    def version(self) -> str:
        """

        Read in the version string from file

        """
        return self.paths.version.read_text().strip()

    @property
    def find(self) -> Callable:
        """

        Use the appropriate package finding function from setuptools

        """
        from fmtr.tools import setup

        if self.paths.is_namespace:
            return setup.find_namespace_packages
        else:
            return setup.find_packages

    @property
    def packages(self) -> List[str]:
        """

        Fetch list of packages excluding canonical paths

        """
        excludes = list(Constants.PACKAGE_EXCLUDE_DIRS) + [f'{name}.*' for name in Constants.PACKAGE_EXCLUDE_DIRS if '*' not in name]
        packages = self.find(where=str(self.paths.repo), exclude=excludes)
        return packages

    @property
    def package_dir(self):
        """

        Needs to be relative apparently as absolute paths break during packaging

        """
        if self.paths.is_namespace:
            return {'': '.'}
        else:
            return None

    @property
    def package_data(self):
        """

        Default package data is just the version file

        """
        return {self.name: [Constants.FILENAME_VERSION]}

    @property
    def url(self) -> str:
        """

        Default to GitHub URL

        """
        return f'https://github.com/{self.org}/{self.name}'

    @property
    def data(self) -> Dict[str, Any]:
        """

        Generate data for use by setuptools

        """
        data = dict(
            name=self.name,
            version=self.version,
            author=self.author,
            author_email=self.AUTHOR_EMAIL,
            url=self.url,
            license=f'Copyright © {datetime.now().year} {self.copyright}. All rights reserved.',
            long_description=self.long_description,
            long_description_content_type='text/markdown',
            packages=self.packages,
            package_dir=self.package_dir,
            package_data=self.package_data,
            entry_points=dict(
                console_scripts=self.console_scripts,
            ),
            install_requires=self.dependencies.install,
            extras_require=self.dependencies.extras,
        ) | self.kwargs
        return data

    def setup(self):
        """

        Call setuptools.setup using generated data

        """

        from fmtr.tools import setup

        return setup.setup_setuptools(**self.data)

    def __repr__(self) -> str:
        """

        Show library name

        """
        return f'{self.__class__.__name__}("{self.name}")'

class Tools:
    """

    Helper for downstream libraries to specify lists of `fmtr.tools` extras

    """
    MASK = f'{Constants.LIBRARY_NAME}[{{extras}}]'

    def __init__(self, *extras):
        self.extras = extras

    def __str__(self):
        extras_str = ','.join(self.extras)
        return self.MASK.format(extras=extras_str)



class Dependencies:
    ALL = 'all'
    INSTALL = 'install'

    def __init__(self, **kwargs):
        self.dependencies = kwargs

    def resolve_values(self, key) -> List[str]:
        """

        Flatten a list of dependencies.

        """
        values_resolved = []
        values = self.dependencies[key]

        for value in values:
            if value == key or value not in self.dependencies:
                # Add the value directly if it references itself or is not a dependency key.
                values_resolved.append(str(value))
            else:
                # Recurse into nested dependencies.
                values_resolved += self.resolve_values(value)

        return values_resolved

    @cached_property
    def extras(self) -> Dict[str, List[str]]:
        """

        Flatten dependencies.

        """
        resolved = {key: self.resolve_values(key) for key in self.dependencies.keys()}
        resolved.pop(self.INSTALL, None)
        resolved[self.ALL] = list(set(chain.from_iterable(resolved.values())))
        return resolved

    @cached_property
    def install(self) -> List[str]:
        """

        Get install_requires

        """
        if self.INSTALL in self.dependencies:
            return self.resolve_values(self.INSTALL)
        else:
            return []


if __name__ == '__main__':
    ...



import flet as ft
from dataclasses import dataclass
from flet.core.gesture_detector import TapEvent
from flet.core.page import Page
from flet.core.types import ColorValue, IconValue
from functools import cached_property
from typing import Optional

from fmtr.tools.logging_tools import logger


class SliderSteps(ft.Slider):
    """

    Slider control using step instead of divisions

    """

    def __init__(self, *args, min=10, max=100, step=10, **kwargs):
        self.step = step
        divisions = (max - min) // step
        super().__init__(*args, min=min, max=max, divisions=divisions, **kwargs)


class Cell(ft.DataCell):
    """

    Context-aware, clickable data cell.

    """

    def __init__(self, series, column):
        """

        Store context

        """
        self.series = series
        self.column = column
        self.value = series[column]

        super().__init__(self.gesture_detector)

    @cached_property
    def text(self):
        """

        Cell contents text

        """

        return ft.Text(str(self.value), color=self.color, bgcolor=self.bgcolor)

    @property
    def color(self):
        """

        Basic conditional formatting

        """
        if self.value is None:
            return ft.Colors.GREY
        else:
            return None

    @property
    def bgcolor(self):
        """

        Basic conditional formatting

        """
        return None

    @cached_property
    def gesture_detector(self):
        """

        Make arbitrary content clickable

        """
        return ft.GestureDetector(content=self.text, on_tap=self.click_tap, on_double_tap=self.click_double_tap)

    async def click_tap(self, event: TapEvent):
        """

        Default cell click behavior — override in subclass if needed

        """
        value = await self.click(event=event)
        event.page.update()
        return value

    async def click(self, event: Optional[TapEvent] = None):
        """

        Default cell click behavior — override in subclass if needed

        """
        logger.info(f"Clicked {self.column=} {self.series.name=} {self.value=}")

    async def click_double_tap(self, event: TapEvent):
        """

        Default cell click behavior — override in subclass if needed

        """
        value = await self.double_click(event=event)
        event.page.update()
        return value

    async def double_click(self, event: Optional[TapEvent] = None):
        """

        Default cell double click behavior — override in subclass if needed

        """
        logger.info(f"Double-clicked {self.column=} {self.series.id=} {self.value=}")


class Row(ft.DataRow):
    """

    Instantiate a row from a series

    """

    TypesCells = {None: Cell}

    def __init__(self, series):
        self.series = series
        super().__init__(self.cells_controls, color=self.row_color)

    @cached_property
    def cells_data(self) -> dict[str, list[Cell]]:
        """

        Cell controls lookup

        """
        data = {}
        for col in self.series.index:
            default = self.TypesCells[None]
            TypeCell = self.TypesCells.get(col, default)
            data.setdefault(col, []).append(TypeCell(self.series, col))
        return data

    @cached_property
    def cells_controls(self) -> list[Cell]:
        """

        Flat list of controls

        """
        controls = []
        for cells in self.cells_data.values():
            for cell in cells:
                controls.append(cell)
        return controls

    def __getitem__(self, item) -> list[Cell]:
        """

        Get cells by column name

        """
        return self.cells_data[item]

    @property
    def row_color(self):
        """

        Basic conditional formatting

        """
        return None


class Column(ft.DataColumn):
    """

    Column stub

    """

    def __init__(self, col_name: str):
        """

        Store context

        """

        self.col_name = col_name

        super().__init__(label=self.text)

    @cached_property
    def text(self):
        """

        Cell contents text

        """
        return ft.Text(str(self.col_name), weight=self.weight)

    @property
    def weight(self):
        """

        Default bold headers

        """
        return ft.FontWeight.BOLD


class Table(ft.DataTable):
    """

    Dataframe with clickable cells

    """

    TypesRows = {None: Row}
    TypesColumns = {None: Column}

    def __init__(self, df):  # todo move to submodule with tabular deps
        """

        Set columns/rows using relevant types

        """
        self.df = df
        super().__init__(columns=self.columns_controls, rows=self.rows_controls)

    @cached_property
    def columns_data(self) -> dict[str, list[Column]]:
        """

        Columns controls lookup

        """
        data = {}
        for col in self.df.columns:
            default = self.TypesColumns[None]
            TypeColumn = self.TypesColumns.get(col, default)
            data.setdefault(col, []).append(TypeColumn(col))
        return data

    @cached_property
    def columns_controls(self) -> list[Column]:
        """

        Flat list of controls

        """
        controls = []
        for columns in self.columns_data.values():
            for column in columns:
                controls.append(column)
        return controls

    @cached_property
    def rows_data(self) -> dict[str, list[Row]]:
        """

        Row controls lookup

        """
        data = {}
        for index, row in self.df.iterrows():
            default = self.TypesRows[None]
            TypeRow = self.TypesRows.get(index, default)
            data.setdefault(index, []).append(TypeRow(row))
        return data

    @cached_property
    def rows_controls(self) -> list[Row]:
        """

        Flat list of controls

        """
        controls = []
        for rows in self.rows_data.values():
            for row in rows:
                controls.append(row)
        return controls

    def __getitem__(self, item) -> list[Row]:
        """

        Get row by index

        """

        return self.rows_data[item]


@dataclass
class NotificationDatum:
    """

    Color and icon for notification bar

    """
    color: ColorValue
    icon: IconValue


class NotificationBar(ft.SnackBar):
    """

    Combined logging and notification bar: infers the appropriate UI representation from the log level

    """
    DATA = {
        logger.info: NotificationDatum(color=ft.Colors.BLUE, icon=ft.Icons.INFO),
        logger.warning: NotificationDatum(color=ft.Colors.AMBER, icon=ft.Icons.WARNING),
        logger.error: NotificationDatum(color=ft.Colors.RED, icon=ft.Icons.ERROR),
        logger.debug: NotificationDatum(color=ft.Colors.GREY, icon=ft.Icons.BUG_REPORT),
        logger.exception: NotificationDatum(color=ft.Colors.RED_ACCENT, icon=ft.Icons.REPORT),
    }

    def __init__(self, msg: str, method=logger.info):
        """

        Log the message immediately, otherwise configure notification bar icon/color.

        """
        self.msg = msg
        self.method = method
        self.method(msg)

        icon = ft.Icon(self.data.icon, color=self.data.color)
        text = ft.Text(self.msg)
        content = ft.Row(controls=[icon, text])
        super().__init__(content=content)

    @cached_property
    def data(self) -> NotificationDatum:
        """

        Fetching data using logging method.

        """
        return self.DATA[self.method]

    def show(self, page: Page):
        """

        Show the notification on the relevant page.

        """
        page.open(self)



import regex as re
from dataclasses import dataclass, asdict
from functools import cached_property
from typing import List, Any

from fmtr.tools import Constants
from fmtr.tools.logging_tools import logger
from fmtr.tools.string_tools import join


class RewriteCircularLoopError(Exception):
    """

    Circular loop error

    """


MASK_GROUP = '(?:{pattern})'
MASK_NAMED = r"(?P<{key}>{pattern})"


def alt(*patterns):
    patterns = sorted(patterns, key=len, reverse=True)
    pattern = '|'.join(patterns)
    pattern = MASK_GROUP.format(pattern=pattern)
    return pattern

@dataclass
class Key:
    RECORD_SEP = '␞'
    FILLS = None

    def flatten(self, data):
        """

        Flatten/serialise dictionary data

        """
        pairs = [f'{value}' for key, value in data.items()]
        string = self.RECORD_SEP.join(pairs)
        return string

    @cached_property
    def pattern(self):
        """

        Serialise to pattern

        """
        data = {
            key:
                MASK_NAMED.format(
                    key=key,
                    pattern=value.format_map(self.fills))
            for key, value in asdict(self).items()
        }
        pattern = self.flatten(data)
        return pattern

    @cached_property
    def rx(self):
        """

        Compile to Regular Expression

        """
        return re.compile(self.pattern)

    @cached_property
    def string(self):
        """

        Serialise to string

        """
        string = self.flatten(asdict(self))
        return string

    @cached_property
    def fills(self):
        """

        Add key names as regex group names

        """
        return {key: MASK_NAMED.format(key=key, pattern=value) for key, value in self.FILLS.items()}


    def transform(self, match: re.Match):
        """

        Transform match object into a new object of the same type.

        """
        fills = match.groupdict()
        data = {key: value.format_map(fills) for key, value in asdict(self).items()}
        obj = self.__class__(**data)
        return obj


@dataclass
class Item:
    """

    Key-value pair

    """
    source: Key
    target: Key


@dataclass(kw_only=True)
class Transformer:
    """
    
    Pattern-based, dictionary-like mapper.
    Compiles an complex set of rules into single regex pattern, and determines which rule matched.
    Inputs are then transformed according to the matching rule.
    Works like a pattern-based dictionary when is_recursive==False.
    Works something like an FSA/transducer when is_recursive=True.

    """
    PREFIX_GROUP = '__'
    items: List[Item]
    default: Any = None
    is_recursive: bool = False

    def __post_init__(self):
        """

        Compile on init

        """
        return self.compile(clear=False)

    def compile(self, clear=True):
        """

        Re/compile regex pattern, invalidating existing caches if recompile.

        """
        if clear:
            del self.pattern
            del self.rx

        with logger.span(f'Compiling expression {len(self.items)=}'):
            rx = self.rx
        logger.debug(f'Compiled successfully {rx.groups=}')

    @cached_property
    def pattern(self) -> str:
        """
        
        Dynamically generated regex pattern based on the rules provided.

        """
        patterns = [
            MASK_NAMED.format(key=f'{self.PREFIX_GROUP}{i}', pattern=item.source.pattern)
            for i, item in enumerate(self.items)
        ]
        pattern = alt(*patterns)
        return pattern

    @cached_property
    def rx(self) -> re.Pattern:
        """

        Regex object.

        """
        return re.compile(self.pattern)

    def get_default(self, key: Key) -> Any:
        """

        Define what to return in case of no match

        """
        if self.is_recursive:
            return key
        else:
            return self.default

    def get(self, key: Key) -> Key | Any:
        """

        Use recursive or single lookup pass, depending on whether recursive lookups have been specified.

        """
        if self.is_recursive:
            with logger.span(f'Transforming recursively {key=}...'):
                return self.get_recursive(key)
        else:
            with logger.span(f'Transforming linearly {key=}...'):
                return self.get_one(key)

    def get_one(self, key: Key) -> Key | Any:
        """

        Single lookup pass.
        Lookup the source string based on the matching rule.

        """

        match = self.rx.fullmatch(key.string)

        if not match:
            value = self.get_default(key)
            logger.debug(f'No match for {key=}. Returning {self.get_default(key)=}')
        else:

            match_ids = {name: v for name, v in match.groupdict().items() if v}
            rule_ids = {
                int(id.removeprefix(self.PREFIX_GROUP))
                for id in match_ids.keys() if id.startswith(self.PREFIX_GROUP)
            }

            if len(rule_ids) != 1:
                msg = f'Multiple group matches: {rule_ids}'
                raise ValueError(msg)

            rule_id = next(iter(rule_ids))
            rule = self.items[rule_id]

            logger.debug(f'Matched using {rule_id=}: {rule.source=}')

            if isinstance(rule.target, Key):
                value = rule.target.transform(match)
            else:
                value = rule.target

            logger.debug(f'Transformed using {rule_id=}: {key=} → {value=}')

        return value

    def get_recursive(self, key: Key) -> Key | Any:
        """

        Lookup the provided key by continuously applying transforms until no changes are made
        or a circular loop is detected.

        """
        history = []
        previous = key

        def get_history_str():
            return join(history, sep=Constants.ARROW_SEP)

        while True:
            if previous in history:
                history.append(previous)
                msg = f'Loop detected on node "{previous}": {get_history_str()}'
                raise RewriteCircularLoopError(msg)

            history.append(previous)
            new = previous
            new = self.get_one(new)
            if new == previous:
                break
            previous = new

            if not isinstance(new, Key):
                history.append(previous)
                break

        if len(history) == 1:
            history_str = 'No transforms performed.'
        else:
            history_str = get_history_str()
        logger.debug(f'Finished transforming: {history_str}')

        return previous


if __name__ == '__main__':
    ...



import logging
import os

from fmtr.tools import environment_tools
from fmtr.tools.constants import Constants

if environment_tools.IS_DEV:
    STREAM_DEFAULT = ENVIRONMENT_DEFAULT = Constants.DEVELOPMENT
else:
    STREAM_DEFAULT = None
    ENVIRONMENT_DEFAULT = Constants.PRODUCTION

IS_DEBUG = environment_tools.get(Constants.FMTR_LOG_LEVEL_KEY, None, converter=str.upper) == 'DEBUG'
LEVEL_DEFAULT = logging.DEBUG if IS_DEBUG else logging.INFO


def null_scrubber(match):
    """

    Effectively disable scrubbing

    """
    return match.value

def get_logger(name, version=None, host=Constants.FMTR_OBS_HOST, key=None, org=Constants.ORG_NAME,
               stream=STREAM_DEFAULT, environment=ENVIRONMENT_DEFAULT, level=LEVEL_DEFAULT):
    """

    Get a pre-configured logfire logger, if dependency is present, otherwise default to native logger.

    """

    stream = stream or name

    try:
        import logfire
    except ImportError:
        logger = logging.getLogger(None)
        logger.setLevel(level)
        logger.warning(f'Logging dependencies not installed. Using native logger.')

        return logger

    logger = logfire

    if key is None:
        key = environment_tools.get(Constants.FMTR_OBS_API_KEY_KEY, default=None)

    if key:
        url = f"https://{host}/api/{org}/v1/traces"
        headers = f"Authorization=Basic {key},stream-name={stream}"

        os.environ["OTEL_EXPORTER_OTLP_TRACES_ENDPOINT"] = url
        os.environ["OTEL_EXPORTER_OTLP_HEADERS"] = headers
        os.environ["OTEL_EXPORTER_OTLP_INSECURE"] = str(False).lower()

    if not version:
        from fmtr.tools import version_tools
        version = version_tools.read()

    # Rigmarole to translate native levels to logfire/otel ones.
    lev_num_otel = logfire._internal.constants.LOGGING_TO_OTEL_LEVEL_NUMBERS[level]
    lev_name_otel = logfire._internal.constants.NUMBER_TO_LEVEL[lev_num_otel]

    console_opts = logfire.ConsoleOptions(
        colors='always',
        min_log_level=lev_name_otel,
    )

    logfire.configure(
        service_name=name,
        service_version=version,
        environment=environment,
        send_to_logfire=False,
        console=console_opts,
        scrubbing=logfire.ScrubbingOptions(callback=null_scrubber)
    )

    if key is None:
        msg = f'Observability dependencies installed, but "{Constants.FMTR_OBS_API_KEY_KEY}" not set. Cloud observability will be disabled.'
        logger.warning(msg)

    return logger


logger = get_logger(name=Constants.LIBRARY_NAME)

if __name__ == '__main__':
    logger.info('Hello World')
    logger.warning('test warning')
    logger.debug('Hello World')
    logger



import httpx
from httpx_retries import RetryTransport

from fmtr.tools import logging_tools

logging_tools.logger.instrument_httpx()


class Client(httpx.Client):
    """

    Instrumented client base

    """

    TRANSPORT = RetryTransport()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, transport=self.TRANSPORT, **kwargs)


client = Client()

if __name__ == '__main__':
    resp = client.get('http://httpbin.org/delay/10')
    resp



from dataclasses import dataclass
from dns import query as dnspython_query, message as dnspython_message, rdatatype as dnspython_rdatatype, rcode as dnspython_rcode
from functools import cached_property
from httpx_retries import Retry, RetryTransport
from typing import Optional

from fmtr.tools import http_tools as http
from fmtr.tools.dns_tools.dm import Exchange, Response
from fmtr.tools.logging_tools import logger

RETRY_STRATEGY = Retry(
    total=2,  # initial + 1 retry
    allowed_methods={"GET", "POST"},
    status_forcelist={502, 503, 504},
    retry_on_exceptions=None,  # defaults to httpx.TransportError etc.
    backoff_factor=0.25,  # short backoff (e.g. 0.25s, 0.5s)
    max_backoff_wait=0.75,  # max total delay before giving up
    backoff_jitter=0.1,  # small jitter to avoid retry bursts
    respect_retry_after_header=False,  # DoH resolvers probably won't set this
)


class HTTPClientDoH(http.Client):
    """

    Base HTTP client for DoH-appropriate retry strategy.

    """
    TRANSPORT = RetryTransport(retry=RETRY_STRATEGY)


@dataclass
class Plain:
    """

    Plain DNS

    """
    host: str
    port: int = 53
    ttl_min: Optional[int] = None

    def resolve(self, exchange: Exchange):

        with logger.span(f'UDP {self.host}:{self.port}'):
            response_plain = dnspython_query.udp(q=exchange.query_last, where=self.host, port=self.port)
            response = Response.from_message(response_plain)
            for answer in response.message.answer:
                answer.ttl = max(answer.ttl, self.ttl_min or answer.ttl)

        exchange.response = response


@dataclass
class HTTP:
    """

    DNS over HTTP

    """

    HEADERS = {"Content-Type": "application/dns-message"}
    CLIENT = HTTPClientDoH()
    BOOTSTRAP = Plain('8.8.8.8')

    host: str
    url: str


    @cached_property
    def ip(self):
        message = dnspython_message.make_query(self.host, dnspython_rdatatype.A, flags=0)
        exchange = Exchange.from_wire(message.to_wire(), ip=None, port=None)
        self.BOOTSTRAP.resolve(exchange)
        ip = next(iter(exchange.response.answer.items.keys())).address
        return ip

    def resolve(self, exchange: Exchange):
        """

        Resolve via DoH

        """

        headers = self.HEADERS | dict(Host=self.host)
        url = self.url.format(host=self.ip)

        try:
            response_doh = self.CLIENT.post(url, headers=headers, content=exchange.query_last.to_wire())
            response_doh.raise_for_status()
            response = Response.from_http(response_doh)
            exchange.response = response

        except Exception as exception:
            exchange.response.message.set_rcode(dnspython_rcode.SERVFAIL)
            exchange.response.is_complete = True
            logger.exception(exception)



from tokenizers import Tokenizer
from tokenizers import normalizers
from tokenizers import pre_tokenizers
from tokenizers.models import BPE
from tokenizers.models import WordLevel
from tokenizers.processors import TemplateProcessing
from tokenizers.trainers import BpeTrainer
from tokenizers.trainers import WordLevelTrainer
from typing import List

from fmtr.tools.path_tools import Path


class TokenConfig:
    """

    Default config for special characters etc.

    """
    PAD = '<pad>'
    UNK = '<unk>'
    CLS = '<cls>'
    BOS = '<bos>'
    EOS = '<eos>'
    SEP = '<sep>'
    MASK = '<mask>'

    COW_PREFIX = '...'
    EOW_SUFFIX = '|'

    SPECIALS = [PAD, UNK, CLS, BOS, EOS, SEP, MASK]

    VARIABLE_LENGTH = -1
    VOCAB_SIZE = 10_000


def fix_length(tokenizer: Tokenizer, length: int):
    """

    Fix tokenizer length if input is too long

    """
    if length > TokenConfig.VARIABLE_LENGTH:
        tokenizer.enable_padding(pad_token=TokenConfig.PAD, pad_id=tokenizer.token_to_id(TokenConfig.PAD),
                                 length=length)
        tokenizer.enable_truncation(max_length=length)

    return tokenizer


def get_template(cls=True, sep=True, pair=False):
    """

    Create a tokenizers template

    """

    if pair:
        n = 1
        letter = 'B'
        template = get_template(cls, sep, pair=False)
    else:
        n = 0
        letter = 'A'
        template = []

    if cls and not pair:
        template.append(f'{TokenConfig.CLS}:{n}')

    template.append(f'${letter}:{n}')

    if sep:
        template.append(f'{TokenConfig.SEP}:{n}')

    return template


def add_template(tokenizer: Tokenizer, cls, sep):
    """

    Add a tokenizers template

    """
    tokenizer.post_processor = TemplateProcessing(
        single=get_template(cls=cls, sep=sep, pair=False),
        pair=get_template(cls=cls, sep=sep, pair=True),
        special_tokens=[
            (token_special, tokenizer.token_to_id(token_special))
            for token_special in TokenConfig.SPECIALS
        ],
    )
    return tokenizer


def train_fixed_vocab_tokenizer(tokens: List[str], length: int = TokenConfig.VARIABLE_LENGTH,
                                show_progress: bool = True, cls: bool = True,
                                sep: bool = True) -> Tokenizer:
    """

    "Train" a word level tokenizer from a fixed, unique set of tokens.

    """
    model = WordLevel(unk_token=TokenConfig.UNK)
    tokenizer = Tokenizer(model)
    tokenizer.normalizer = normalizers.Lowercase()
    trainer = WordLevelTrainer(min_frequency=0, special_tokens=TokenConfig.SPECIALS)
    tokenizer.train_from_iterator([sorted(set(tokens))], trainer=trainer, show_progress=show_progress)
    fix_length(tokenizer, length)
    add_template(tokenizer, cls, sep)
    return tokenizer


def train_bpe_tokenizer(data, vocab_size: int = TokenConfig.VOCAB_SIZE, min_frequency: int = 0,
                        length: int = TokenConfig.VARIABLE_LENGTH,
                        show_progress: bool = True, cls: bool = True, sep: bool = True) -> Tokenizer:
    """

    Train byte-pair encoder from given parameters

    """
    model = BPE(unk_token=TokenConfig.UNK)
    tokenizer = Tokenizer(model)

    normalizers_bpe = [normalizers.NFC(), normalizers.Strip(), normalizers.StripAccents(), normalizers.Lowercase()]
    tokenizer.normalizer = normalizers.Sequence(normalizers_bpe)

    pre_tokenizers_bpe = [pre_tokenizers.Whitespace(), pre_tokenizers.Digits(individual_digits=True)]
    tokenizer.pre_tokenizer = pre_tokenizers.Sequence(pre_tokenizers_bpe)

    trainer = BpeTrainer(
        special_tokens=TokenConfig.SPECIALS,
        vocab_size=vocab_size,
        min_frequency=min_frequency,
        end_of_word_suffix=TokenConfig.EOW_SUFFIX,
        continuing_subword_prefix=TokenConfig.COW_PREFIX,
        show_progress=show_progress,
        initial_alphabet=pre_tokenizers.ByteLevel.alphabet()
    )

    is_data_files = isinstance(data, list) and all(isinstance(datum, Path) for datum in data)
    if is_data_files:
        tokenizer.train([str(path) for path in data], trainer)
    else:
        tokenizer.train_from_iterator(data, trainer)

    fix_length(tokenizer, length)
    add_template(tokenizer, cls, sep)

    return tokenizer



import functools
import inspect
from typing import Tuple, Callable, Self

from fmtr.tools import context_tools


def combine_args_kwargs(args: dict=None, kwargs: dict=None) -> dict:
    """

    Combines arguments and keyword arguments into a single dictionary.

    """
    args = args or []
    kwargs = kwargs or {}
    args = {i: arg for i, arg in enumerate(args)}
    args.update(kwargs)
    if all(isinstance(key, int) for key in args.keys()):
        args = list(args.values())
    return args


def split_args_kwargs(args_kwargs: dict) -> Tuple[list, dict]:
    """

    Splits arguments and keyword arguments into a list and a dictionary.

    """
    if isinstance(args_kwargs, list):
        args, kwargs = args_kwargs, {}
    else:
        args = [arg for key, arg in args_kwargs.items() if isinstance(key, int)]
        kwargs = {key: arg for key, arg in args_kwargs.items() if not isinstance(key, int)}

    return args, kwargs


class MethodDecorator:
    """

    Bound method decorator with overridable start/stop and context manager

    """

    CONTEXT_KEY = 'context'

    def __init__(self):
        """

        Initialise the decorator itself with any arguments

        """
        self.func = None

    def __call__(self, func: Callable) -> Self:
        """

        Add the (unbound) method.

        """
        self.func = func
        functools.update_wrapper(self, func)
        return self

    def __get__(self, instance, owner):
        """

        Wrap bound method at runtime, call start/stop within context.

        """
        if instance is None:  # Class method called.
            return self.func

        if inspect.iscoroutinefunction(self.func):
            async def async_wrapper(*args, **kwargs):
                with self.get_context(instance):
                    self.start(instance, *args, **kwargs)
                    result = await self.func(instance, *args, **kwargs)
                    self.stop(instance, *args, **kwargs)
                    return result

            return async_wrapper

        else:
            def sync_wrapper(*args, **kwargs):
                with self.get_context(instance):
                    self.start(instance, *args, **kwargs)
                    result = self.func(instance, *args, **kwargs)
                    self.stop(instance, *args, **kwargs)
                    return result

            return sync_wrapper

    def get_context(self, instance):
        """

        If the instance has a context attribute, use that - otherwise use a null context.

        """
        context = getattr(instance, self.CONTEXT_KEY, None)
        if context:
            return context
        return context_tools.null()

    def start(self, instance, *args, **kwargs):
        pass

    def stop(self, instance, *args, **kwargs):
        pass



import os
from typing import TypeVar, Generic, Type

import flet as ft
from flet.core.control_event import ControlEvent
from flet.core.types import AppView
from flet.core.view import View

from fmtr.tools import environment_tools
from fmtr.tools.constants import Constants
from fmtr.tools.function_tools import MethodDecorator
from fmtr.tools.interface_tools.context import Context
from fmtr.tools.logging_tools import logger


class update(MethodDecorator):
    """

    Update the page after the decorated function is called.

    """

    def stop(self, instance, *args, **kwargs):
        instance.page.update()


class progress(update):
    """

    Run the function while a progress indicator (e.g. spinner) is and within the object-defined context (e.g. logging span).

    """

    def start(self, instance, *args, **kwargs):
        """

        Make progress visible and update.

        """
        instance.progress.visible = True
        instance.page.update()

    def stop(self, instance, *args, **kwargs):
        """

         Make progress not visible and update.

        """
        instance.progress.visible = False
        super().stop(instance)


T = TypeVar('T', bound=Context)


class Base(Generic[T], ft.Column):
    """

    Simple interface base class.

    """
    TITLE = 'Base Interface'
    HOST = '0.0.0.0'
    PORT = 8080
    URL = Constants.FMTR_DEV_INTERFACE_URL if environment_tools.IS_DEV else None
    APPVIEW = AppView.WEB_BROWSER
    PATH_ASSETS = None
    PATH_UPLOADS = None
    SCROLL = ft.ScrollMode.AUTO

    SECRET_KEY_KEY = 'FLET_SECRET_KEY'
    ROUTE_ROOT = '/'

    TypeContext: Type[T] = Context

    @classmethod
    async def new(cls, page: ft.Page):
        """

        Interface entry point/async constructor. Set relevant callbacks, and add instantiated self to page views.

        Override this to work with `Context`, do async setup. Otherwise, override __init__ (which is regular Column __init__) for a simple interface.

        """
        page.scroll = cls.SCROLL
        page.title = cls.TITLE
        page.on_connect = cls.on_connect
        page.on_disconnect = cls.on_disconnect
        page.on_route_change = cls.route
        page.on_view_pop = cls.pop
        page.theme = cls.get_theme()

        context = cls.TypeContext(page=page)
        self = cls()
        self.context = context

        page.controls.append(self)
        page.update()

        return self

    @classmethod
    def route(cls, event: ft.RouteChangeEvent):
        """

        Overridable router.

        """
        logger.debug(f'Route change: {event=}')

    @classmethod
    def pop(cls, view: View, page: ft.Page):
        """

        Overridable view pop.

        """
        logger.debug(f'View popped: {page.route=} {len(page.views)=} {view=}')

    @classmethod
    def on_connect(cls, event: ControlEvent):
        """

        Log connections

        """
        page = event.control
        logger.warning(f'Connect: {page.client_user_agent=} {page.platform.name=}')

    @classmethod
    def on_disconnect(cls, event: ControlEvent):
        """

        Log disconnections

        """
        page = event.control
        logger.warning(f'Disconnect {page.client_user_agent=} {page.platform.name=}')


    @classmethod
    def get_theme(self):
        """

        Overridable theme definition

        """
        text_style = ft.TextStyle(size=20)
        theme = ft.Theme(
            text_theme=ft.TextTheme(body_large=text_style),
        )
        return theme

    @classmethod
    def launch(cls):
        """

        Launch via async constructor method

        """

        if cls.URL:
            url = cls.URL
        else:
            url = f'http://{cls.HOST}:{cls.PORT}'

        if not environment_tools.get(cls.SECRET_KEY_KEY, default=None):
            os.environ["FLET_SECRET_KEY"] = os.urandom(12).hex()

        logger.info(f"Launching {cls.TITLE} at {url}")
        ft.app(cls.new, view=cls.APPVIEW, host=cls.HOST, port=cls.PORT, assets_dir=cls.PATH_ASSETS, upload_dir=cls.PATH_UPLOADS)


class Test(Base[Context]):
    """

    Simple test interface, showing typing example.

    """
    TypeContext: Type[Context] = Context

    TITLE = 'Test Interface'

    def __init__(self):
        controls = [ft.Text(self.TITLE)]
        super().__init__(controls=controls)

if __name__ == "__main__":
    Test.launch()



from fmtr.tools import yaml_tools
from fmtr.tools.tests.helpers import SERIALIZATION_DATA


def test_yaml():
    """

    Simple YAML round trip test

    """
    expected = SERIALIZATION_DATA
    actual = yaml_tools.from_yaml(yaml_tools.to_yaml(expected))
    assert actual == expected
